import json
import logging

import google_auth_oauthlib.flow
import google.oauth2.credentials
import google.oauth2.service_account
import googleapiclient.discovery

# Create logger object
logger = logging.getLogger(__name__)


class GoogleAPI:
    """
    Helper class for creating Google APIs resources.

    The class avails two classmethods
    :py:meth:`from_service_account_file` and
    :py:meth:`from_authorized_user_file` for creating
    an instance of the :py:class:`GoogleAPI` class.

    Parameters
    ----------
    credentials : google.oauth2.credentials.Credentials
        The Credentials object used for authentication
        when building API resources.
    api_name : str
        The name of the API endpoint.
    api_version : str
        The version of the API endpoint.
    """

    def __init__(
            self,
            credentials: google.oauth2.credentials.Credentials,
            api_name: str,
            api_version: str) -> None:
        self._credentials = credentials
        self._api_name = api_name
        self._api_version = api_version
        self.resource = googleapiclient.discovery.build(
            serviceName=self._api_name,
            version=self._api_version,
            credentials=self._credentials,
            cache_discovery=False,
        )

    @classmethod
    def from_service_account_file(
            cls,
            api_name: str,
            api_version: str,
            credential_filename: str,
            credential_scopes: list):
        """
        Creates an instance of :py:class:`GoogleAPI` from a
        service account file.

        This is the recommended approach for authenticating
        with Google APIs.

        Parameters
        ----------
        api_name : str
            The name of the API endpoint.
        api_version : str
            The version of the API endpoint.
        credential_filename : str
            The name of the service account file.
        credential_scopes : list
            List of OAuth 2.0 Scopes for Google APIs.

        Returns
        -------
        GoogleAPI
            Returns an instance of the :py:class:`GoogleAPI`
            class.
        """

        # Generate credentials object
        credentials = google.oauth2.service_account.Credentials.from_service_account_file(
            filename=credential_filename,
            scopes=credential_scopes,
        )
        return GoogleAPI(
            credentials=credentials,
            api_name=api_name,
            api_version=api_version,
        )

    @classmethod
    def from_authorized_user_file(
            cls,
            api_name: str,
            api_version: str,
            credential_filename: str,
            credential_scopes: list):
        """
        Creates an instance of :py:class:`GoogleAPI` from an
        authorized user file.

        This classmethod explicitly uses the authorized user file
        created by the :py:func:`generate_authorized_user_file`
        function. Alternatively, you may also create your own as
        long as the required keys `client_id`, `client_secret`,
        and `refresh_token` is present in the JSON file.

        Parameters
        ----------
        api_name : str
            The name of the API endpoint.
        api_version : str
            The version of the API endpoint.
        credential_filename : str
            The name of the authorized user file.
        credential_scopes : list
            List of OAuth 2.0 Scopes for Google APIs.

        Returns
        -------
        GoogleAPI
            Returns an instance of the :py:class:`GoogleAPI`
            class.
        """

        # Generate credentials object
        credentials = google.oauth2.credentials.Credentials.from_authorized_user_file(
            filename=credential_filename,
            scopes=credential_scopes,
        )
        return GoogleAPI(
            credentials=credentials,
            api_name=api_name,
            api_version=api_version,
        )


def generate_authorized_user_file(
        credential_filename: str,
        credential_scopes: list) -> str:
    """
    Helper function for generating authorized user file.

    This function uses the
    :py:class:`google_auth_oauthlib.flow.InstalledAppFlow` class
    to retrieve generate OAuth access and refresh tokens
    from a `client secrets file`_, which are then saved
    to a JSON authorized user file.

    The keys required to be written to this authorized
    user file are those required by the classmethod
    :py:meth:`google.oauth2.credentials.Credentials.from_authorized_user_file`.

    The OAuth 2.0 `access scopes`_ depends on the APIs you
    intend to access, and they will incrementally granted on
    top of the existing granted scopes.

    .. _`client secrets file`:
        https://developers.google.com/identity/protocols/OAuth2WebServer#creatingcred
    .. _`access scopes`:
        https://developers.google.com/identity/protocols/oauth2/scopes

    Note
    ----
    When generating a new OAuth 2.0 Client ID in Google Cloud,
    ensure that the application type used is set to ``Other``.

    Parameters
    ----------
    credential_filename : str
        The name of the client secret file.
    credential_scopes : list
        List of OAuth 2.0 Scopes for Google APIs.

    Returns
    -------
    str
        The name of the JSON authorized user file.

    References
    ----------
    https://google-auth.readthedocs.io/en/latest/index.html
    https://google-auth-oauthlib.readthedocs.io/en/latest/index.html
    """

    # Use the client_secret.json file to identify the application requesting
    # authorization. The client ID (from that file) and access scopes are required.
    flow = google_auth_oauthlib.flow.InstalledAppFlow.from_client_secrets_file(
        client_secrets_file=credential_filename,
        scopes=credential_scopes,
    )
    credentials = flow.run_console(
        # Enable offline access so that you can refresh an access token without
        # re-prompting the user for permission. Recommended for web server apps.
        access_type="offline",
        # Enable incremental authorization. Recommended as a best practice.
        include_granted_scopes="true",
    )

    # Create JSON data
    data = {
        "client_id": credentials.client_id,
        "client_secret": credentials.client_secret,
        "refresh_token": credentials.refresh_token,
    }

    # Write JSON data to file
    filename = f"authorized_user_{credentials.client_id}.json"
    with open(filename, "x") as f:
        json.dump(data, f)
    return filename
