from __future__ import annotations

import os
import logging
import json
from functools import wraps

# Create logger object
logger = logging.getLogger(__name__)


def persist_state(filename: str, keys: list[str]):
    """
    Decorator method for persisting state for each pipeline stage.

    This decorator restores state, if available, from a state file
    before a stage is ran and exports the state back to the file
    thereafter.

    The primary use case of using this decorator is to support
    cross-stage communication where pipeline stages are executed
    sequentially albeit from different class instances. By specifying
    the list of keys, the decorator will attempt to set the values
    of these keys from the state file as instance attributes before
    executing the stage method, and export them back into the state
    file thereafter to be used by subsequent pipeline stages, if any.

    Parameters
    ----------
    filename : str
        The name of the JSON-encoded file.
    keys : list[str]
        The list of keys to restore to and export from the state file.
    """

    def decorator(method):
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            logger.debug(f"List of persistent keys (others will be ignored): {keys}")

            # Restore state if available
            if os.path.isfile(filename):
                logger.info(f"State file is present, restoring from {filename}.")
                with open(filename, "r") as f:
                    data = json.load(f)
                logger.debug(f"Restoring state from file: {data}")
                for key in keys:
                    if key in data:
                        value = data[key]
                        logger.debug(f"Setting self.{key} = {value}.")
                        setattr(self, key, value)
                    else:
                        logger.warning(f"Key '{key}' is not present in state file.")

            # Calls wrapped method
            output = method(self, *args, **kwargs)

            # Export state to file
            data = {}  # Should export a new state file, not overwrite!
            for key in keys:
                if hasattr(self, key):
                    data[key] = getattr(self, key)
                else:
                    logger.warning(f"Key '{key}' is not assigned in instance.")
            logger.debug(f"Exporting state to file: {data}")
            with open(filename, "w") as f:
                json.dump(data, f)

            return output
        return wrapper
    return decorator
