import logging

from google.cloud import bigquery
from google.cloud.exceptions import GoogleCloudError

# Create logger object
logger = logging.getLogger(__name__)


def load_to_bigquery(
        client: bigquery.Client,
        dataset: str,
        table: str,
        uri: str,
        partition_field: str,
        clustering_fields: list = None,
        schema: list = None) -> None:
    """
    Loads a Cloud Storage object into a BigQuery partition.

    This function allows for highly consistent data loads
    into a ``DATE`` partition without the need of running
    an additional deduplication query.

    However, in order for the data load to be consistent,
    the partition date must representatively match the
    data it loads. Deviations from this will cause data
    to be placed in wrong partitions and result in,
    potentially, erroneous queries.

    Parameters
    ----------
    client : google.cloud.bigquery.client.Client
        The Google BigQuery client object.
    dataset : str
        The name of the destination dataset.
    table : str
        The name of the destination table.
    uri : str
        The URI or glob pattern of Cloud Storage object.
    partition_field : str
        The name of the column to be used as the
        parition_field.
    clustering_fields : list, optional
        The list of column names, in sequence, to be
        used a clustering fields.
    schema : list, optional
        The list of schema fields, in sequence, to be
        used when loading into the BigQuery table.
    """

    # Set job configuration
    logger.info("Generating job configuration")
    job_config = {
        # Create table if it does not already exists
        "create_disposition": bigquery.job.CreateDisposition.CREATE_IF_NEEDED,

        # Overrides table even if it already exists
        "write_disposition": bigquery.job.WriteDisposition.WRITE_TRUNCATE,

        # Enforces date-level partitioning.
        "time_partitioning": bigquery.TimePartitioning(
            field=partition_field,
        ),

        # Set clustering fields if present.
        "clustering_fields": clustering_fields,

        # Set schema if present.
        "schema": schema,

        # Ensures that the header row will not get parsed by the schema.
        "skip_leading_rows": 1,

        # NOTE: Setting autodetect=True is permanent for the lifecycle
        # for a BigQuery table. Some configuration parameters may also
        # be ignored when this is true, leading to unexpected behaviours.
        # "autodetect": True,
    }
    logger.debug(f"Job configuration: {job_config}")

    # Load to BigQuery
    logger.info("Starting BigQuery Job.")
    # NOTE: `bigquery.client.Client.dataset()` is deprecated in 1.24.0
    # https://github.com/googleapis/google-cloud-python/pull/7753
    table_ref = bigquery.TableReference.from_string(
        f"{client.project}.{dataset}.{table}",
    )
    job = client.load_table_from_uri(
        source_uris=uri,
        destination=table_ref,
        job_config=bigquery.LoadJobConfig(**job_config),
    )

    # Waits for job to finish
    try:
        logger.info("Waiting for BigQuery Job to complete.")
        job.result()
    except GoogleCloudError:
        logger.critical("Encountered BigQuery loading errors!")
        for error in job.errors:
            logger.critical(error)
        raise

    # Update job status
    logger.info(f"Current BigQuery Job status: {job.state}")
    if job.state == "DONE":
        logger.info(f"Successfully uploaded '{uri}' to '{dataset}.{table}' on Google BigQuery.")


def update_table(
        client: bigquery.Client,
        dataset: str,
        table: str,
        **kwargs) -> None:
    """
    Helper function to update table attributes.

    Parameters
    ----------
    client : google.cloud.bigquery.client.Client
        The Google BigQuery client object.
    dataset : str
        The name of the destination dataset.
    table : str
        The name of the destination table.
    **kwargs
        Key-value mapping of
        :py:class:`google.cloud.bigquery.table.Table`
        attributes. If a key is not a valid
        attribute of
        :py:class:`google.cloud.bigquery.table.Table`
        or if value is :py:obj:`None`, then the
        key-value pair will be ignored.
    """

    # Get table object
    t = bigquery.Table.from_string(
        f"{client.project}.{dataset}.{table}",
    )

    # Set attributes and execute update
    fields = []
    for k, v in kwargs.items():
        if hasattr(t, k):
            logger.debug(f"Setting Table.{k}={v}")
            setattr(t, k, v)
            fields.append(k)
    logger.debug(f"Table attributes to be updated: {fields}")
    client.update_table(
        table=t,
        fields=fields,
    )
