import os


def get_file_size(filename: str) -> str:
    """
    Returns a human-readable expression of a file size.

    Parameters
    ----------
    filename : str
        The name of the file.

    Returns
    -------
    str:
        Returns a human-readable expression of a file size.
    """

    # Get file size
    size = os.path.getsize(filename)
    for unit in ["B", "KiB", "MiB", "GiB", "TiB"]:
        if size < 1024:
            break
        size /= 1024
    return f"{size:.2f} {unit}"


def get_row_matching_prefix(
        filename: str,
        prefix: str,
        reverse: bool = False) -> int:
    """
    Returns the number of the first matching row that starts
    with a prefix string.

    If ``reverse=True``, the row number is returned as if the
    file is looped in reverse order.

    This function loops through every row in the file from
    the top and returns the row number where a row begins
    with a specified prefix. If no matches were found, the
    integer zero is returned.

    Parameters
    ----------
    filename : str
        The name of the file.
    prefix : str
        The prefix string to be matched.
    reverse : bool
        Whether to return the row number as if looped in reverse
        order. Defaults to False.

    Returns
    -------
    int:
        Returns the row number that starts with a prefix string.
        Else, returns 0 if there are no matches.
    """

    with open(filename, "r") as f:
        c = 0  # Count of rows
        n = 0  # Row number
        for i, row in enumerate(f):
            c += 1
            if row.startswith(prefix):
                n = c
                # Return first matched row if reverse=False
                if not reverse:
                    return n
        # Return last match if a match is found
        if n:
            return c - (n - 1)
    # Return 0 if no match is found
    return n
