import math
import json
import logging


class GKELoggingHandler(logging.StreamHandler):
    """
    Handler to format log messages as expected by the `fluentd`_
    plugin in Google Kubernetes Engine, so that metadata such
    as log level are properly set.

    Adapted from :py:class:`google.cloud.logging.handlers.container_engine.ContainerEngineHandler`.

    .. _`fluentd`:
        https://www.fluentd.org

    Parameters
    ----------
    name : str, optional
        The name of the custom log in Stackdriver Logging.
    stream : io.IOBase, optional
        The stream to be used by the handler.
    """

    def __init__(self, name=None, stream=None):
        super().__init__(stream=stream)
        self.name = name

    def format(self, record):
        """
        Formats the message into JSON expected by the `default
        configuration`_ of the `fluentd`_ plugin in Google
        Kubernetes Engine.

        .. _`default configuration`:
            https://cloud.google.com/logging/docs/agent/configuration#default-config
        .. _`fluentd`:
            https://www.fluentd.org

        Parameters
        ----------
        record : logging.LogRecord
            The log record.

        Returns
        -------
        str
            A JSON string formatted for GKE fluentd.
        """
        message = super().format(record)
        subsecond, second = math.modf(record.created)
        payload = {
            # NOTE: Some fields are handled specially by the logging agent
            # https://cloud.google.com/logging/docs/agent/configuration#special-fields
            "message": message,
            "timestamp": {"seconds": int(second), "nanos": int(subsecond * 1e9)},
            "thread": record.thread,
            "severity": record.levelname,
            # Other optional fields available in `logging.LogRecord`
            # https://docs.python.org/3/library/logging.html#logging.LogRecord
            "name": record.name,
            "pathname": record.pathname,
            "lineno": record.lineno,
            # "msg": record.msg,
            # "args": record.args,
            # "exc_info": record.exc_info,
            # "func": record.func,
            # "sinfo": record.sinfo,
        }
        return json.dumps(payload)
