import datetime
import logging
import uuid
import csv
import os
import json

import pytz
from google.cloud import bigquery
from google.cloud import storage

from konekuta.utils import (
    get_reference_date,
    get_relative_date_range,
    validate_date_ranges,
    parse_timestamp,
)
from konekuta.inspectors import (
    get_file_size,
)
from konekuta.transformers import (
    split_data,
)
from konekuta.schema import (
    sync_schema,
)
from konekuta.helpers.storage import (
    upload_to_storage,
)
from konekuta.helpers.bigquery import (
    load_to_bigquery,
)

# Create logger object
logger = logging.getLogger(__name__)


class Pipeline:
    """
    A base class for constructing ELT data pipelines.

    Parameters
    ----------
    gcs_bucket : str
        The name of the Google Cloud Storage bucket.
    gcs_directory : str
        The name of the Google Cloud Storage directory prefix.
        This name will be used as a prefix for all files uploaded.
    bq_dataset : str
        The name of the Google BigQuery dataset.
    bq_table : str
        The name of the Google BigQuery table.
    bq_schema : list, optional
        The schema of the source file and destination file. This
        schema defined here will take precedence over any schema
        defined using the ``schema_filename`` argument.
    bq_partitioning_field : str, optional
        The partitioning field to be used when uploading the data
        into Google BigQuery.
    bq_clustering_fields : list, optional
        The list of clustering fields, in order, to be used when
        uploading the data into Google BigQuery.
    gcloud_labels : dict, optional
        The labels to be included in all Google Cloud resources
        interactions where possible.
    scheduled_timestamp : str, optional
        The timezone-aware timestamp when the pipeline is intended
        to be ran. If absent, the ``scheduled_timestamp`` is assumed
        to be the current timestamp.
    start_date : str, optional
        The start date (YYYY-MM-DD) used when querying data from
        a data source.
    end_date : str, optional
        The end date (YYYY-MM-DD) used when querying data from
        a data source.
    timezone : str, optional
        The timezone used for deriving the reference date when start
        and end date ranges are not supplied or are invalid.
        Defaults to ``UTC``.
    lookback : int, optional
        The lookback window, in days, used for deriving the start and
        end date ranges when they are not supplied or are invalid.
        Defaults to ``1``.
    offset : int, optional
        The date offset, in days, used for deriving the start and
        end date ranges when they are not supplied or are invalid.
        Defaults to ``1``.
    schema_filename : str, optional
        The name of the schema file that, if specified, will be
        used as the reference schema. Consider using this argument
        for standardised schema fields and using the ``bq_schema``
        argument to add or override any other fields.
    strict : bool, optional
        Whether to execute the pipeline in strict mode. If True,
        some errors, though non-fatal, will cause the pipeline
        to exit with error.
        Defaults to :py:obj:`False`.
    debug : bool, optional
        Whether to execute the pipeline in debug mode. If True,
        additional debug information will be logged.
        Defaults to :py:obj:`False`.
    """

    def __init__(
            self,
            gcs_bucket: str,
            gcs_directory: str,
            bq_dataset: str,
            bq_table: str,
            bq_schema: list = None,
            bq_partitioning_field: str = None,
            bq_clustering_fields: list = None,
            gcloud_labels: dict = None,
            scheduled_timestamp: str = None,
            start_date: str = None,
            end_date: str = None,
            timezone: str = "UTC",
            lookback: int = 1,
            offset: int = 1,
            schema_filename: str = None,
            strict: bool = False,
            debug: bool = False) -> None:

        # Assign variables as instance attributes
        self.gcs_bucket = gcs_bucket
        self.gcs_directory = gcs_directory
        self.bq_dataset = bq_dataset
        self.bq_table = bq_table
        self.bq_schema = bq_schema
        self.bq_partitioning_field = bq_partitioning_field
        self.bq_clustering_fields = bq_clustering_fields
        self.gcloud_labels = gcloud_labels
        self.scheduled_timestamp = scheduled_timestamp
        self.start_date = start_date
        self.end_date = end_date
        self.timezone = timezone
        self.lookback = lookback
        self.offset = offset
        self.schema_filename = schema_filename
        self.strict = strict
        self.debug = debug

        # Instantiate filenames
        self.raw_filenames = None
        self.working_filenames = None
        self.transformed_filenames = None
        self._headers = None
        self._reference_schema = {}
        self._custom_schema = {}
        self._accumulated_schema = None
        self._synchronised_schema = None

        # Set instance identifiers
        self.instance_timestamp = datetime.datetime.now(pytz.utc)
        logger.debug(f"Instance Timestamp: {self.instance_timestamp}")
        self.instance_id = uuid.uuid4()
        logger.debug(f"Instance ID: {self.instance_id}")

        # Parse scheduled timestamp
        logger.info("Checking if 'scheduled_timestamp' argument is provided.")
        if not self.scheduled_timestamp:
            logger.warning("Argument 'scheduled_timestamp' was not provided, using the current timestamp instead.")
            self.scheduled_timestamp = self.instance_timestamp
        else:
            self.scheduled_timestamp = parse_timestamp(
                timestamp=self.scheduled_timestamp,
            )
        logger.debug(f"Scheduled timestamp: {self.scheduled_timestamp}")

        # Override start_date and end_date if either are not specified
        # TODO: Add tests
        logger.info("Checking if 'start_date' and 'end_date' arguments are supplied.")
        if not self.start_date or not self.end_date:
            logger.warning("Either 'start_date' or 'end_date' was not supplied!")

            # Set reference date
            logger.info("Getting reference date from current timestamp and timezone.")
            self.reference_date = get_reference_date(
                timestamp=self.scheduled_timestamp,
                timezone=self.timezone,
            )
            logger.debug(f"Localised Reference Date: {self.reference_date}")

            # Get date ranges
            logger.info(f"Deriving date ranges with lookback window of {self.lookback} day(s) and date offset of {self.offset} day(s).")
            self.start_date, self.end_date = get_relative_date_range(
                date=self.reference_date,
                lookback=self.lookback,
                offset=self.offset,
            )
            logger.debug(f"Derived date ranges: {self.start_date} -> {self.end_date}")

        # Validate date ranges if available
        logger.info(f"Validating date ranges: {self.start_date} -> {self.end_date}")
        valid = validate_date_ranges(
            start_date=self.start_date,
            end_date=self.end_date,
        )
        if not valid:
            raise ValueError(f"Supplied date range arguments were invalid: {self.start_date} -> {self.end_date}")

        # Import reference schema if specified
        # TODO: Consider also importing from GCS?
        if self.schema_filename:
            logger.info(f"Importing reference schema from '{self.schema_filename}'.")
            with open(self.schema_filename, "r") as f:
                schema = json.load(f)
                # Try loading new schema format first
                # NOTE: The preferred schema format is to have
                # schema fields populated in the `fields` key
                # of the JSON document.
                try:
                    key = "fields"
                    self._reference_schema = schema[key]
                # Fallback to legacy schema format
                # TODO: Once we deprecate list-like schema input, we can omit TypeError
                except (TypeError, KeyError):
                    logger.warning(f"Key '{key}' was not found in the schema file '{self.schema_filename}', importing entire JSON document as-is.")
                    self._reference_schema = schema
        else:
            # TODO: Add tests
            logger.warning("Argument 'schema_filename' is empty, which means that schema WILL NOT be dynamically mapped.")

        # Assign custom schema if provided
        # TODO: Should we deprecate this field?
        if self.bq_schema:
            self._custom_schema = self.bq_schema

        # The following uses credentials inferred from the local environment
        # using Application Default Credentials.
        # https://googleapis.dev/python/google-api-core/latest/auth.html
        self.bigquery_client = bigquery.Client()
        self.storage_client = storage.Client()

    def run(self):
        """Executes the full ETL data pipeline."""

        # Calls ETL methods
        self.extract()
        self.transform()
        self.load()

        # Calls clean method
        self.clean()

    def extract(self):
        """
        Extracts raw data from a data source.

        The :py:meth:`extract()` method expects the ``self.raw_filenames``
        attribute to contain a list of filenames. Each file should
        be a raw, untouched data file that was downloaded from a
        data source.

        These files will be uploaded to Cloud Storage as-is and
        stored in the ``/raw/`` sub-directory.
        """

        # Handle if self.raw_filenames is None
        logger.debug("Checking if instance attribute 'raw_filenames' was assigned.")
        if not self.raw_filenames:
            # TODO: Custom exception
            raise Exception("Instance attribute 'raw_filenames' was not assigned!")

        # Upload raw file to Google Cloud
        logger.debug(f"Uploading files in queue: {self.raw_filenames}")
        for filename in self.raw_filenames:
            # Log file size
            size = get_file_size(filename)
            logger.debug(f"File '{filename}' has size: {size}")

            # Upload
            upload_to_storage(
                client=self.storage_client,
                source_filename=filename,
                bucket_name=self.gcs_bucket,
                blob_name=f"{self.gcs_directory}/raw/{self.instance_timestamp.astimezone(pytz.utc).date().isoformat()}-{self.instance_id}/{filename}",
            )

    def transform(self):
        """
        Transforms raw data into the desired ingestion format.

        The :py:meth:`transform()` method expects the ``self.working_filenames``
        attribute to contain a list of filenames. Each file should be
        an **uncompressed, utf-8 encoded, comma-separated** file. Each
        file should also contain the **same set of column headers in the
        first row** with **no leading or trailing metadata rows**. The
        ``DATE``-like column should contain dates formatted in ISO8601
        ``YYYY-MM-DD`` format.

        If the ``self.working_filenames`` attribute is not explicitly
        assigned, it shall reference the filenames directly from
        ``self.raw_filenames``.

        In instances where the raw files do not meet the above criteria,
        it is necessary to subclass this method to transform the raw files
        before calling this method via the ``super()`` function.

        The files found in ``self.working_filenames`` will be split in
        multiple date-partitioned data files. These files will be uploaded
        to Cloud Storage and stored in the ``/transformed/`` sub-directory.

        References
        ----------
        https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types
        """

        # Handle if self.working_filenames is None
        logger.debug("Checking if instance attribute 'working_filenames' was assigned.")
        if not self.working_filenames:
            logger.warning(f"Instance attribute 'working_filenames' was not assigned, falling back to 'raw_filenames': {self.raw_filenames}")
            self.working_filenames = self.raw_filenames

        # Read column headers
        # TODO: Can this be delegated to another file-reading function?
        # TODO: This currently reads headers of the first file, which may or may not be a desirable behaviour.
        logger.debug("Reading column headers for schema mapping later.")
        first = self.working_filenames[0]
        with open(first, "r") as f:
            reader = csv.DictReader(f)
            self._headers = reader.fieldnames

        # Split data file into multiple rows
        logger.debug(f"Splitting files into multiple date-partitioned CSV files using the '{self.bq_partitioning_field}' column.")
        self.transformed_filenames = split_data(
            filenames=self.working_filenames,
            fieldname=self.bq_partitioning_field,
        )

        # Upload all partitioned files to GCS
        logger.debug(f"Uploading files in queue: {self.transformed_filenames}")
        for filename in self.transformed_filenames:
            upload_to_storage(
                client=self.storage_client,
                source_filename=filename,
                bucket_name=self.gcs_bucket,
                blob_name=f"{self.gcs_directory}/transformed/{filename}",
            )

    def load(self):
        """
        Loads transformed data from Cloud Storage to BigQuery.

        The :py:meth:`load()` method expects the relevant schema indexes
        to be imported using the ``schema_filename`` keyword argument
        when instantiating the class.

        Column headers, which were read from one of the files earlier,
        are matched to the schema index. The schema is built from
        the headers in an ordinal manner, and later passed to BigQuery
        when loading the files from the directory.

        Since BigQuery load jobs are free, all files matching
        ``/transformed/*.csv`` are always freshly loaded into the table
        with the current schema. If the schema does not match the underlying
        data, this may lead to data integrity issues even if no errors
        were raised. Any data in an existing BigQuery table will be
        overwritten by the data files matching the above glob pattern.
        """

        # Accumulate schema
        logger.debug("Accumulating reference and custom schema.")
        # TODO: What if custom schema format is wrong?
        # TODO: Add a custom schema validation function
        self._accumulated_schema = {
            # Unpack the reference schema first
            **self._reference_schema,
            # Unpack any custom schema (overrides the above)
            **self._custom_schema,
        }
        # NOTE: Disabled debug logs for accumulated schema because the output is too big
        # logger.debug(f"Accumulated schema: {self._accumulated_schema}")

        # Synchronise schema
        logger.debug("Synchronising schema with actual file headers.")
        self._synchronised_schema = sync_schema(
            # This should be a combination of all schema
            schema=self._accumulated_schema,
            headers=self._headers,
        )
        logger.debug(f"Synchronised schema: {self._synchronised_schema}")

        # Load all files to BQ
        uri_pattern = f"gs://{self.gcs_bucket}/{self.gcs_directory}/transformed/*.csv"
        load_to_bigquery(
            client=self.bigquery_client,
            dataset=self.bq_dataset,
            table=self.bq_table,
            uri=uri_pattern,
            partition_field=self.bq_partitioning_field,
            schema=self._synchronised_schema,
        )

    def clean(self):
        """
        Helper method to remove all runtime files.

        The method checks for filenames in the following
        attributes and remove any files with the exact
        filename if it exists:

        - ``raw_filenames``
        - ``working_filenames``
        - ``transformed_filenames``

        If any created files are not appended to any
        lists in the above attributes, these files will
        be invisible to this method.
        """

        # Clean up
        attributes = (
            "raw_filenames",
            "working_filenames",
            "transformed_filenames",
        )
        for attr in attributes:
            # Ensures that AttributeError exception is caught
            filenames = getattr(self, attr, None)
            if filenames:
                for filename in filenames:
                    if os.path.isfile(filename):
                        os.remove(filename)
