import logging
import re

# Create logger object
logger = logging.getLogger(__name__)

# Auto-generated field descriptiion
AUTO_GENERATED_FIELD_DESCRIPTION = "No match found."


def sync_schema(
        schema: dict,
        headers: list) -> dict:
    """
    Returns synchronised schema in `BigQuery schema format`_.

    This function performs the logic of synchronising
    a list of headers against a schema map. For any
    header in the ``headers`` argument that is found in
    the schema provided in the ``schema`` argument, the
    schema output will be derived directly. For those
    that are not, they will default to ``NULLABLE`` mode
    with ``STRING`` data type.

    The outcome of this function ensures that the schema
    output will match both the **count** and **order** of
    the column headers, which is an important requirement
    for error-free data load process.

    Since the function relies heavily on the schema
    argument to derive the correct schema type and mode,
    it is important to ensure the schema index is as extensive
    and accurate as possible.

    .. _`BigQuery schema format`:
        https://cloud.google.com/bigquery/docs/schemas#creating_a_json_schema_file

    Parameters
    ----------
    schema : dict
        The schema, in :py:obj:`dict` format, containing
        an index of fields and its schema format in a nested
        ``spec`` key.
    headers : list
        The list of column headers in :py:obj:``str`` format.

    Returns
    -------
    dict
        Returns the synchronised schema in BigQuery
        schema format.
    """

    # Validate and prepare existing schema
    # NOTE: Only `name` and `type` is a required field by BigQuery
    # https://cloud.google.com/bigquery/docs/schemas#specifying_a_json_schema_file
    required_keys = (
        # "description",
        "name",
        "type",
        # "mode",
    )

    # Build schema
    # TODO: Attempt to guess type; use a flag
    validated = []
    for header in headers:
        # Check if field exists
        try:
            field = schema.get(header)
        except AttributeError:
            raise TypeError(f"Schema does not appear to be a 'dict' object: {schema}")
        if not field:
            logger.warning(f"Header {header} has no known schema type; defaulting as NULLABLE STRING.")
            field = {
                "spec": {
                    # TODO: This header should be sanitised right?
                    "name": header,
                    "description": AUTO_GENERATED_FIELD_DESCRIPTION,
                    "type": "STRING",
                    "mode": "NULLABLE",
                },
            }

        # Validate required keys
        spec = field["spec"]
        for key in required_keys:
            if key not in spec:
                raise KeyError(f"Required keys {required_keys} was not found in field spec: {spec}.")

        # Sanitise header
        sanitised = sanitise_header(spec["name"])
        spec["name"] = sanitised

        # Append to validated schema
        validated.append(spec)
    return validated


def sanitise_header(header: str) -> str:
    """
    Returns a sanitised string that matches the
    ``[A-Za-z_][A-Za-z0-9_]{0,127}`` regex pattern.

    This function will attempt to sanitise the
    header by converting illegal characters into
    underscores, which is the default behaviour in
    BigQuery.

    A valid string is only returned if:
        * length of header is no more than 128 characters.
        * header does not start with reserved prefixes.
        * header starts with a letter or underscore.

    Parameters
    ----------
    header : str
        The name of the column header.

    Returns
    -------
    str
        The sanitised header if it is valid, else
        ``None``.

    Examples
    --------
    >>> sanitise_header(" this.is-a_header")
    '_this_is_a_header'
    """

    # Check if header is within length limit
    if len(header) > 128:
        logger.error(f"Tried to sanitise {header} but it has >128 characters")
        return

    # Sanitise header
    sanitised = re.sub(r"\W", "_", header)

    # Check if header starts with any of the reserved prefixes
    reserved_prefixes = (
        "_TABLE_",
        "_FILE_",
        "_PARTITION",
    )
    if sanitised.startswith(reserved_prefixes):
        logger.error(f"Sanitised {header} to {sanitised} but it starts with one of BigQuery's reserved prefixes: {reserved_prefixes}")
        return

    # Check if sanitised string starts with illegal character
    # NOTE: This logic is needed because the re.sub pattern
    # will not replace numbers at the start of the string.
    # TODO: Consider integrating the rule into re.sub pattern.
    if not re.match(r"[a-zA-Z_]", sanitised[0]):
        logger.error(f"Sanitised {header} to {sanitised} but it should start with a letter or an underscore")
        return

    return sanitised
