import csv
import os
import datetime
import logging
import subprocess
import platform

# Create logger object
logger = logging.getLogger(__name__)


def split_data(
        filenames: list,
        fieldname: str = "Date",
        pattern: str = "%Y-%m-%d",
        format_dates: bool = False) -> list:
    """
    Splits a date-level level dataset into multiple files.

    This function is meant to support a date-partitioned
    date warehouse design. Regardless of the date pattern
    in the input file, the name of the output file will
    always follow the ISO 8601 date format.

    Parameters
    ----------
    filenames : list
        The list of source file names.
    fieldname : str
        The name of column header containing date strings.
        Defaults to ``Date``.
    pattern : str, optional
        The pattern used to interpret the string-formatted
        dates. Defaults to ``%Y-%m-%d``.
    format_dates : bool, optional
        Formats date in each row as ISO8601 ``%Y-%m-%d``
        format. Defaults to :py:obj:`False`.

    Returns
    -------
    list
        The list of names of files created by the function.
    """

    # Initialise variable for tracking files created by function
    files = set()

    # Initialise variable for tracking headers
    headers = None

    # Process each file
    for filename in filenames:
        logger.debug(f"Splitting '{filename}'")
        with open(filename, "r") as fi:
            reader = csv.DictReader(fi)

            # Check if headers is same as previously encountered
            if headers:
                if headers != reader.fieldnames:
                    raise ValueError(f"The current file headers does not match headers previously encountered files: {headers} != {reader.fieldnames}")
            else:
                headers = reader.fieldnames

            # Loop each row in file
            for row in reader:
                # Open corresponding file
                d = datetime.datetime.strptime(row[fieldname], pattern).date()
                df = f"{d.isoformat()}.csv"
                mode = "x"

                # Check if file exists before
                created = df in files
                exists = os.path.isfile(df)
                if exists and created:
                    mode = "a"

                # Open file
                # TODO: Consider processing in chunks to avoid opening/closing file for each row
                with open(df, mode=mode) as fo:
                    writer = csv.DictWriter(fo, fieldnames=reader.fieldnames)

                    # Write header if file does not exists before
                    if not exists:
                        writer.writeheader()
                        files.add(df)

                    # Format date if desired
                    if format_dates:
                        row[fieldname] = d.isoformat()

                    # Write row data
                    writer.writerow(row)

    # Log error if no files were generated
    generated = list(files)
    if not generated:
        logger.error(f"Expected to generate at least 1 file but there was none: {generated}")

    # Return set as list
    return generated


def remove_leading_rows(
        filename: str,
        rows: int) -> None:
    """
    Removes n leading rows from a file.

    This command is logically similar to the following ``sed``
    command:

    .. code-block:: shell

        sed -i -e "1,${ROWS}d" ${FILENAME}

    Note
    ----
    This function uses the :py:mod:`subprocess` module
    to call the GNU ``sed`` command, which is then used to
    apply remove rows from the target file. In environments
    where ``sed`` is not available, this function may
    behave unexpectedly.

    Parameters
    ----------
    filename : str
        The name of the file.
    rows : int
        The number of leading rows to remove from a file.
    """

    # Return if rows value is invalid
    if rows < 1:
        logger.warning("No rows were removed since rows < 1")
        return

    # Set commands based on system
    os = platform.system()
    command = ["sed", "-i", "-e", f"1,{rows}d", f"{filename}"]
    if os == "Darwin":
        command = ["sed", "-i", "", "-E", f"1,{rows}d", f"{filename}"]

    # Calls command and wait till command finishes
    try:
        subprocess.run(
            command,
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        logger.critical(f"Command stdout: {e.stdout}")
        logger.critical(f"Command stderr: {e.stderr}")
        raise


def remove_trailing_rows(
        filename: str,
        rows: int) -> None:
    """
    Removes n trailing rows from a file.

    This command is logically similar to the following ``sed``
    command:

    .. code-block:: shell

        sed -i -e ":a" -e "$d;N;1,${ROWS}ba" -e "P;D" ${FILENAME}

    Parameters
    ----------
    filename : str
        The name of the file.
    rows : int
        The number of trailing rows to remove from a file.
    """

    # Return if rows value is invalid
    if rows < 1:
        logger.warning("No rows were removed since rows < 1")
        return

    # Remove trailing lines
    with open(filename, "rb+") as f:
        # Moves pointer to end of file
        f.seek(0, os.SEEK_END)

        lines = 0
        while f.tell() > 0:
            # Moves pointer backwards from current position
            f.seek(-1, os.SEEK_CUR)

            # Increment line count if character is LF
            char = f.read(1)
            if char == b"\n":
                lines += 1

            # Truncate file if line endings count are met
            # NOTE: The last line ending must be taken
            # into account since the `lines` variable
            # is a count of line endings from the back.
            if lines == rows + 1:
                f.truncate()
                break

            # Moves pointer backwards from current position
            f.seek(-1, os.SEEK_CUR)
