from __future__ import annotations

import datetime
import logging
import socket

import pytz

# Create logger object
logger = logging.getLogger(__name__)


def parse_timestamp(timestamp: str) -> datetime.datetime:
    """
    Returns a timezone-aware timestamp object.

    Parameters
    ----------
    timestamp : str
        The timestamp in ISO8601 :py:obj:`str` format.
        If the input timestamp is timezone-naive, then
        it is assumed to be in UTC timezone.

    Returns
    -------
    datetime.datetime
        Returns a timezone-aware timestamp object.
    """

    # Type checking
    if not isinstance(timestamp, str):
        raise TypeError(f"Expected argument 'timestamp' to be an instance of str object: {timestamp} ({type(timestamp)})")

    # Parse timestamp
    ts = datetime.datetime.fromisoformat(timestamp)

    # Check if timestamp is timezone-aware
    if not ts.tzinfo:
        logger.warning(f"Timestamp '{ts}' is timezone-naive, setting timezone to '{pytz.utc}'.")
        ts = ts.replace(tzinfo=pytz.utc)

    # Normalise timestamp
    n = pytz.utc.normalize(ts)
    return n


def get_reference_date(
        timestamp: datetime.datetime,
        timezone: str) -> datetime.date:
    """
    Returns a localised reference date based on a supplied
    timezone.

    The ``timestamp`` variable supplied to the method should be
    timezone aware.

    Parameters
    ----------
    timestamp : datetime.datetime
        The timezone-aware timestamp to be used to derive the
        reference date.
    timezone : str
        The timezone used to localise the timestamp.

    Returns
    -------
    datetime.date
        Returns the date of a localised timestamp.
    """

    # Type checking
    if not isinstance(timestamp, datetime.datetime):
        raise TypeError(f"Expected argument 'timestamp' to be an instance of datetime.datetime object: {timestamp} ({type(timestamp)})")
    if not isinstance(timezone, str):
        raise TypeError(f"Expected argument 'timezone' to be an instance of str object: {timezone} ({type(timezone)})")

    # Check if timestamp is timezone aware
    if not timestamp.tzinfo:
        raise ValueError("Expected argument 'timestamp' to be timezone-aware")

    # Validate timezone
    logger.info("Parsing timezone parameter.")
    if timezone not in pytz.all_timezones:
        logger.warning(f"Timezone '{timezone}' was not found in pytz.all_timezones, defaulting to 'UTC'.")
        timezone = "UTC"

    # Localise timestamp
    logger.info(f"Localising timestamp {timestamp} to timezone {timezone}")
    tz = pytz.timezone(timezone)
    localised = timestamp.astimezone(tz)
    logger.debug(f"Localised Timestamp: {localised}")

    # Return localised timestamp as date
    d = localised.date()
    return d


def get_relative_date_range(
        date: datetime.date,
        lookback: int = 1,
        offset: int = 1) -> tuple[str, str]:
    """
    Returns a set of date range based on the reference date, lookback
    window, and date offset.

    This function only accepts date objects so it is expected for dates
    to be resolved in the correct timezones before passing into this
    function.

    Parameters
    ----------
    date : datetime.date
        The date to be used as a reference point, which is usually the
        current date.
    lookback : int, optional
        The lookback window in days. A lookback window of 1 day will
        generate a date range of the same day. Defaults to ``1``.
    offset : int, optional
        The date offset in days. A date offset of 1 day will generate
        a date range with an ending date of ``date`` - 1 day. Defaults
        to ``1``.

    Returns
    -------
    tuple[str, str]
        Returns a :py:obj:`tuple` of date ranges in ``YYYY-MM-DD``
        :py:obj:`str` format.

    Examples
    --------
    >>> date = datetime.date(1970, 1, 1)
    >>> # Yesterday
    >>> get_relative_date_range(date)
    ('1969-12-31', '1969-12-31')
    >>> # 7 days ago
    >>> get_relative_date_range(date, offset=7)
    ('1969-12-25', '1969-12-25')
    >>> # Last 7 days
    >>> get_relative_date_range(date, lookback=7)
    ('1969-12-25', '1969-12-31')
    >>> # Last 30 days
    >>> get_relative_date_range(date, lookback=30)
    ('1969-12-02', '1969-12-31')
    """

    # Type checking
    if not isinstance(date, datetime.date):
        raise TypeError(f"Expected argument 'date' to be an instance of datetime.date object: {date} ({type(date)})")
    if not isinstance(lookback, int):
        raise TypeError(f"Expected argument 'lookback' to be an instance of int object: {lookback} ({type(lookback)})")
    if not isinstance(offset, int):
        raise TypeError(f"Expected argument 'offset' to be an instance of int object: {offset} ({type(offset)})")

    # Set lookback to default if out of range
    if lookback < 1:
        default_lookback = 1
        logger.warning(f"The specified lookback period of {lookback} day(s) is invalid, defaulting to lookback period of {default_lookback}")
        lookback = default_lookback

    # Set offset to default if out of range
    if offset < 1:
        default_offset = 1
        logger.warning(f"The specified offset period of {offset} day(s) is invalid, defaulting to offset period of {default_offset}")
        offset = default_offset

    # Generate date ranges
    start = date - datetime.timedelta(days=offset) - datetime.timedelta(days=lookback - 1)
    end = date - datetime.timedelta(days=offset)

    # Return tuple of date ranges
    return start.isoformat(), end.isoformat()


def get_unix_timestamp(
        date: str,
        timezone: str,
        pattern: str = "%Y-%m-%d",
        milliseconds: bool = False) -> int:
    """
    Returns the number of seconds since the Unix epoch.

    Since ``date`` is expected to be timezone-naive, the
    ``timezone`` parameter is required to hydrate the date
    with the desired timezone. The Unix timestamp can then
    be derived from the timezone-aware date.

    Parameters
    ----------
    date : str
        The date in :py:obj:`string` format.
    timezone : str
        The timezone of the localised date.
    pattern : str, optional
        The pattern used to interpret the string-formatted
        dates. Defaults to ``%Y-%m-%d``.
    milliseconds : bool, optional
        Whether to return the timestamp in millisecond accuracy.
        Defaults to :py:obj:`False`.

    Returns
    -------
    int:
        Returns the number of milliseconds from the Unix epoch.
    """

    # Parse date
    d = datetime.datetime.strptime(date, pattern)

    # Hydrate datetime object with timezone
    tz = pytz.timezone(timezone)
    # NOTE: The pytz.localize() method must be preferred over
    # the datetime.replace() method when using the pytz library.
    # d = d.replace(tzinfo=tz)
    d = tz.localize(d)

    # Derive unix timestamp and format in milliseconds
    ts = int(d.timestamp())

    # Return in milliseconds if desired
    if milliseconds:
        return ts * 1000

    return ts


def validate_date_ranges(
        start_date: str,
        end_date: str,
        pattern: str = "%Y-%m-%d") -> bool:
    """
    Returns :py:obj:`True` if date ranges are valid.

    This function checks if date ranges follow the
    ``YYYY-MM-DD`` format and if ``start_date`` is
    on/before ``end_date``.

    Parameters
    ----------
    start_date : str
        The starting date as per format defined in
        the ``pattern`` argument.
    end_date : str
        The ending date as per format defined in
        the ``pattern`` argument.
    pattern : str, optional
        The pattern used to interpret the string-formatted
        dates. Defaults to ``%Y-%m-%d``.

    Returns
    -------
    bool:
        Returns :py:obj:`True` if date ranges are valid.
    """

    # Parse date ranges
    try:
        start = datetime.datetime.strptime(start_date, pattern).date()
        end = datetime.datetime.strptime(end_date, pattern).date()
    except ValueError:
        logger.error(
            f"Date ranges does not conform to pattern: {pattern}!",
            exc_info=True,
        )
        return False

    # Validate date ranges
    if start > end:
        logger.error(f"Starting date '{start}' is after ending date '{end}'!")
        return False
    return True


def is_google_compute_instance():
    """
    Returns True if system is a Google Compute instance.

    This function uses the :py:mod:`socket` module to
    resolve the Google Compute `Metadata Server`_ hostname
    (``metadata.google.internal``) via DNS lookup. If
    the hostname is resolvable, it is assumed that the
    runtime of this function is executed in a Google
    Compute instance.

    .. _`Metadata Server`:
        https://cloud.google.com/compute/docs/storing-retrieving-metadata

    Returns
    -------
    bool
        Returns True if system is a Google Compute instance.
    """

    hostname = "metadata.google.internal"
    try:
        socket.gethostbyname(hostname)
    # NOTE: socket.gaierror is a subclass of OSError
    # but tests are mysteriously failing claiming that
    # it does not inherit BaseException, hence OSError
    # is used here as the exception to be caught.
    except OSError:
        return False
    return True
